from oarepo_model_builder.datatypes import StringDataType

from .field import RegularMarshmallowComponent


class StringMarshmallowComponent(RegularMarshmallowComponent):
    eligible_datatypes = [StringDataType]

    def prepare(self, datatype, *, context, **kwargs):
        min_length = datatype.definition.get("minLength", None)
        max_length = datatype.definition.get("maxLength", None)
        if min_length is None and max_length is None:
            return
        m = datatype.definition.setdefault("marshmallow", {})
        validators = m.setdefault("validators", [])
        args = []
        if min_length is not None:
            args.append("min=%s" % min_length)
        if max_length is not None:
            args.append("max=%s" % max_length)
        validators.append("ma.validate.Length(%s)" % ", ".join(args))



from marshmallow import fields

from .datatypes import DataType


class NumberDataType(DataType):
    facets = {
        "facet-class": "invenio_records_resources.services.records.facets.TermsFacet",
    }


class IntegerDataType(NumberDataType):
    model_type = "integer"

    ui = {
        "marshmallow": {
            "field-class": "marshmallow.fields{ma_fields.Integer}",
        }
    }
    marshmallow = {
        "field-class": "marshmallow.fields{ma_fields.Integer}",
    }
    json_schema = {"type": "integer"}

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Integer(required=False)
        exclusiveMinimum = fields.Integer(required=False)
        maximum = fields.Integer(required=False)
        exclusiveMaximum = fields.Integer(required=False)
        enum = fields.List(fields.Integer(), required=False)


class FloatDataType(NumberDataType):
    model_type = "float"

    ui = {
        "marshmallow": {
            "field-class": "marshmallow.fields{ma_fields.Float}",  # NOSONAR
        }
    }
    marshmallow = {
        "field-class": "marshmallow.fields{ma_fields.Float}",
    }
    json_schema = {"type": "number"}

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Float(required=False)
        exclusiveMinimum = fields.Float(required=False)
        maximum = fields.Float(required=False)
        exclusiveMaximum = fields.Float(required=False)
        enum = fields.List(fields.Float(), required=False)


class DoubleDataType(NumberDataType):
    model_type = "double"

    ui = {
        "marshmallow": {
            "field-class": "marshmallow.fields{ma_fields.Float}",
        }
    }
    marshmallow = {
        "field-class": "marshmallow.fields{ma_fields.Float}",
    }
    json_schema = {"type": "number"}

    class ModelSchema(DataType.ModelSchema):
        minimum = fields.Float(required=False)
        exclusiveMinimum = fields.Float(required=False)
        maximum = fields.Float(required=False)
        exclusiveMaximum = fields.Float(required=False)
        enum = fields.List(fields.Float(), required=False)


class BooleanDataType(DataType):
    model_type = "boolean"

    ui = {
        "marshmallow": {
            "field-class": "marshmallow.fields{ma_fields.Boolean}",
        }
    }
    marshmallow = {
        "field-class": "marshmallow.fields{ma_fields.Boolean}",
    }
    json_schema = {"type": "boolean"}
    facets = {
        "facet-class": "invenio_records_resources.services.records.facets.TermsFacet",
    }



from oarepo_model_builder.schema import ModelSchema


def extract_extended_record(included_data, *, context, **kwargs):
    """
    This function extracts the record to be extended. The extended part is always object
        * If it is passed the whole model file, that is, "record" is inside, return that element
        * If there are 'properties' inside, return the whole object
        * As a fallback, return the whole object

    In all cases, it will strip everything apart from "type" and "properties"
    and keep those in "other_properties" inside the context so that other processors
    can add some of those selectively.
    """
    if (
        "record" in included_data
        and "properties" in included_data["record"]
        and "properties" not in included_data
    ):
        extended_object = included_data["record"]
    elif "properties" in included_data:
        extended_object = included_data
    else:
        context["props"] = included_data
        # need to return shallow copy, as there might be a manipulation with the
        # element and context would be corrupted
        return {**included_data}

    context["props"] = extended_object
    return {
        "properties": extended_object.get("properties", {}),
    }

def add_record_to_included_data(included_data, *, context, **kwargs):
    if "props" in context and "record" in context["props"]:
        included_data["record"] = context["props"]["record"]
    return included_data


def extend_modify_marshmallow(included_data, *, context, **kwargs):
    """
    This processor moves the marshmallow section of the base record to base-class-marshmallow
    and base-class-ui-marshmallow. It also sets the from-base-class flag to True.
    """

    def mark_as_from_base_class(node):
        ret = {**node}
        node_properties = ret.pop("properties", None)
        node_items = ret.pop("items", None)
        ret["from-base-class"] = True

        if node_properties:
            properties = ret.setdefault("properties", {})
            for k, v in node_properties.items():
                v = mark_as_from_base_class(v)
                properties[k] = v
        if node_items:
            ret["items"] = mark_as_from_base_class(node.item)
        ret["base-class-marshmallow"] = ret.pop("marshmallow", {})
        ret["base-class-ui-marshmallow"] = ret.setdefault("ui", {}).pop(
            "marshmallow", {}
        )
        return ret

    def as_array(x):
        if isinstance(x, list):
            return x
        if not x:
            return []
        return [x]

    def replace_use_with_extend(data):
        if isinstance(data, dict):
            if ModelSchema.USE_KEYWORD in data:
                data.setdefault(ModelSchema.EXTEND_KEYWORD, []).extend(
                    as_array(data.pop(ModelSchema.USE_KEYWORD))
                )
            if ModelSchema.REF_KEYWORD in data:
                data.setdefault(ModelSchema.EXTEND_KEYWORD, []).extend(
                    as_array(data.pop(ModelSchema.REF_KEYWORD))
                )
            for v in data.values():
                if isinstance(v, (dict, list)):
                    replace_use_with_extend(v)
        elif isinstance(data, list):
            for v in data:
                if isinstance(v, (dict, list)):
                    replace_use_with_extend(v)

    included_data["marshmallow"] = context["props"].get("marshmallow", {})
    included_data["ui"] = context["props"].get("ui", {})
    ret = mark_as_from_base_class(included_data)

    for ext in (
        ModelSchema.EXTEND_KEYWORD,
        ModelSchema.REF_KEYWORD,
        ModelSchema.USE_KEYWORD,
    ):
        if ext in context["props"]:
            ret[ext] = context["props"][ext]

    replace_use_with_extend(ret)
    return ret


def post_extend_modify_marshmallow(*, element, **kwargs):
    def convert_schema_classes(node):
        node_properties = node.get("properties", None)
        node_items = node.get("items", None)

        was_inherited = "from-base-class" in node
        if not was_inherited:
            return False

        contains_only_inherited_properties = node.pop("from-base-class", False)
        if node_properties:
            for k, v in node_properties.items():
                prop_contains_only_inherited_properties = convert_schema_classes(v)
                if not prop_contains_only_inherited_properties:
                    contains_only_inherited_properties = False
        elif node_items:
            contains_only_inherited_properties = (
                convert_schema_classes(node_items)
                and contains_only_inherited_properties
            )
        base_class_marshmallow = node.pop("base-class-marshmallow", {})
        base_class_ui_marshmallow = node.pop("base-class-ui-marshmallow", {})

        def update_marshmallow(new_marshmallow, base_marshmallow):
            if new_marshmallow.get("generate", True) is False:
                # the class is set to not generate -> if there is a class, do not change it,
                # if not, set it to the base class
                if not new_marshmallow.get("class") and base_marshmallow.get("class"):
                    new_marshmallow["class"] = base_marshmallow["class"]
                return

            if "items" in node:
                # array itself does not have a marshmallow, so no need to modify this
                _update_non_existing(new_marshmallow, base_marshmallow)
                return

            if "properties" not in node:
                # primitive data type -> set it not to be generated unless the field says otherwise
                if "read" not in new_marshmallow:
                    new_marshmallow["read"] = False
                if "write" not in new_marshmallow:
                    new_marshmallow["write"] = False
                for k, v in base_marshmallow.items():
                    if k not in new_marshmallow:
                        new_marshmallow[k] = v
                return

            # now we have an object to modify - convert to base classes only if there are extra properties
            convert_to_base_classes = (
                node_properties and not contains_only_inherited_properties
            )

            if "class" in new_marshmallow:
                # someone added class to the new_marshmallow, so we do not want to change it
                convert_to_base_classes = True

            if convert_to_base_classes:
                if base_marshmallow.get("class"):
                    new_marshmallow["base-classes"] = [base_marshmallow["class"]]
                new_marshmallow["generate"] = True

            elif contains_only_inherited_properties:
                # keep the base class marshmallow, but do not generate the class as it has been generated
                # in the extended library
                new_marshmallow.clear()
                new_marshmallow.update(base_marshmallow)
                new_marshmallow["generate"] = False

            else:
                _update_non_existing(new_marshmallow, base_marshmallow)

        update_marshmallow(node.setdefault("marshmallow", {}), base_class_marshmallow)
        update_marshmallow(
            node.setdefault("ui", {}).setdefault("marshmallow", {}),
            base_class_ui_marshmallow,
        )

        return contains_only_inherited_properties

    convert_schema_classes(element)


def _update_non_existing(target, source):
    for k, v in source.items():
        if k not in target:
            target[k] = v



import marshmallow as ma

from oarepo_model_builder.datatypes import ModelDataType
from oarepo_model_builder.datatypes.components.model.utils import set_default

from ..facets import FacetDefinition
from ..facets.object import ObjectFacetsComponent
from .defaults import DefaultsModelComponent


class FacetsSchema(ma.Schema):
    class Meta:
        unknown = ma.RAISE

    searchable = ma.fields.Bool(required=False)

    extra_code = ma.fields.String(
        attribute="extra-code",
        data_key="extra-code",
        metadata={"doc": "Extra code to be pasted to search options module"},
    )
    module = ma.fields.String(
        metadata={"doc": "Module where the facets will be placed"}
    )
    generate = ma.fields.Boolean()
    skip = ma.fields.Boolean()
    facet_groups = ma.fields.Dict(
        attribute="facet-groups",
        data_key="facet-groups",
        keys=ma.fields.String(),
        values=ma.fields.Dict(keys=ma.fields.String(), values=ma.fields.Integer()),
        metadata={
            "doc": "Groups of facets in the form of {group_name: {facet_path: priority}}. Will merge with facet "
            "groups declared on the individual fields."
        },
    )


class FacetsModelComponent(ObjectFacetsComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [DefaultsModelComponent]

    class ModelSchema(ma.Schema):
        facets = ma.fields.Nested(
            FacetsSchema,
            required=False,
        )

    def before_model_prepare(self, datatype, *, context, **__kwargs):
        module = datatype.definition["module"]["qualified"]
        profile_module = context["profile_module"]

        facets = set_default(datatype, "facets", {})
        facets.setdefault("generate", True)
        facets.setdefault("module", f"{module}.services.{profile_module}.facets")

        facets.setdefault("extra-code", "")
        facets.setdefault("groups", True)

    def build_facet_definition(
        self,
        datatype,
        facet_definition: FacetDefinition,
    ):
        if facet_definition.searchable is None:
            facet_definition.searchable = datatype.definition.get("searchable", True)
        if (
            facet_definition.searchable is not False
            and facet_definition.facet is not False
        ):
            return [facet_definition]
        else:
            # facet will not be generated
            return []



import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.datatypes.components.model.utils import set_default


class ExtResourceSchema(ma.Schema):
    generate = ma.fields.Bool()
    skip = ma.fields.Bool()

    class Meta:
        unknown = ma.RAISE


class ExtResourceModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]

    class ModelSchema(ma.Schema):
        ext_resource = ma.fields.Nested(
            ExtResourceSchema,
            attribute="ext-resource",
            data_key="ext-resource",
        )

    def process_ext_resource(self, datatype, section, **kwargs):
        if datatype.profile == "record":
            cfg = section.config
            cfg["ext-service-name"] = "service_records"
            cfg["ext-resource-name"] = "resource_records"

    def before_model_prepare(self, datatype, *, context, **kwargs):
        if not datatype.profile == "record":
            return
        ext = set_default(datatype, "ext-resource", {})

        ext.setdefault("generate", True)
        ext.setdefault("skip", False)



"""
Datatypes supported out of the box:

integer, float, double:
  constraints: minimumExclusive, maximumExclusive, minimum, maximum

boolean:
  no constraints supported

date, time, datetime
  no constraints supported

edtf, edtf-time, edtf-time-interval, edtf-interval
  no constraints supported

fulltext, keyword, fulltext+keyword:
  minLength, maxLength, pattern

flatten:
  object represented as flatten type in elasticsearch
"""

from .containers import (  # noqa
    ArrayDataType,
    FlatObjectDataType,
    NestedDataType,
    ObjectDataType,
)
from .datatypes import DataType, DataTypeComponent, Section, datatypes  # noqa
from .dates import (  # noqa
    DateDataType,
    DateTimeDataType,
    EDTFDataType,
    EDTFIntervalType,
    EDTFTimeDataType,
    EDTFTimeIntervalType,
    TimeDataType,
)
from .model import ModelDataType  # noqa
from .primitive_types import NumberDataType  # noqa , just for export
from .primitive_types import (
    BooleanDataType,
    DoubleDataType,
    FloatDataType,
    IntegerDataType,
)
from .strings import StringDataType  # noqa , just for export
from .strings import (  # noqa
    FulltextDataType,
    FulltextKeywordDataType,
    HtmlDataType,
    KeywordDataType,
    URLDataType,
    UUIDDataType,
)

DEFAULT_DATATYPES = [
    IntegerDataType,
    FloatDataType,
    DoubleDataType,
    BooleanDataType,
    DateDataType,
    TimeDataType,
    DateTimeDataType,
    EDTFDataType,
    EDTFIntervalType,
    EDTFTimeIntervalType,
    EDTFTimeDataType,
    FulltextDataType,
    KeywordDataType,
    FulltextKeywordDataType,
    UUIDDataType,
    ObjectDataType,
    NestedDataType,
    FlatObjectDataType,
    ArrayDataType,
    URLDataType,
    ModelDataType,
    HtmlDataType,
]



from ..datatypes import Section
from ..utils.dict import dict_get
from .json_base import JSONBaseBuilder


class JSONSchemaBuilder(JSONBaseBuilder):
    TYPE = "jsonschema"
    output_file_type = "jsonschema"
    output_file_name = ["json-schema-settings", "file"]
    parent_module_root_name = "jsonschemas"
    create_parent_packages = True

    def build_node(self, node):
        skip = dict_get(
            self.current_model.definition, ["json-schema-settings", "skip"], False
        )
        if skip:
            return
        generated = self.generate(node)
        self.output.merge(generated)

    def generate(self, node):
        json_schema: Section = node.section_json_schema
        ret = {**json_schema.config}

        if json_schema.children:
            properties = ret.setdefault("properties", {})
            for k, v in json_schema.children.items():
                v = self.generate(v)
                properties[k] = v
        if json_schema.item:
            ret["items"] = self.generate(json_schema.item)
        return ret



import json


def set_default(datatype, *attrs):
    keys = attrs[:-1]
    value = attrs[-1]
    d = datatype.definition
    for k in keys[:-1]:
        if k not in d:
            d[k] = {}
        d = d[k]
    if keys[-1] not in d:
        if callable(value):
            value = value(datatype.definition)
        d[keys[-1]] = value
        return value
    return d[keys[-1]]


def array_contains_value(arr, value):
    if isinstance(value, (int, float, bool, str)):
        return value in arr
    pv = json.dumps(value, sort_keys=True)
    for vv in arr:
        vv = json.dumps(vv, sort_keys=True)
        if vv == pv:
            return True
    return False


def deep_equals(value, value2):
    if isinstance(value, (int, float, bool, str)):
        return value == value2
    return json.dumps(value, sort_keys=True) == json.dumps(value2, sort_keys=True)


def append_array(datatype, *attrs):
    arr = set_default(datatype, *attrs[:-1], [])
    value = attrs[-1]
    if callable(value):
        value = value(datatype.definition)
    if not array_contains_value(arr, value):
        arr.append(value)


def prepend_array(datatype, *attrs):
    arr = set_default(datatype, *attrs[:-1], [])
    value = attrs[-1]
    if callable(value):
        value = value(datatype.definition)
    if not array_contains_value(arr, value):
        arr.insert(0, value)


def place_after(datatype, *attrs):
    value = attrs[-1]
    placeholder = attrs[-2]
    arr = set_default(datatype, *attrs[:-2], [])
    if callable(value):
        value = value(datatype.definition)
    if callable(placeholder):
        placeholder = placeholder(datatype.definition)
    if not array_contains_value(arr, value):
        for idx, elem in enumerate(arr):
            if deep_equals(placeholder, elem):
                arr.insert(idx + 1, value)
                break



from ... import NumberDataType
from .field import RegularMarshmallowComponent


class NumberMarshmallowComponent(RegularMarshmallowComponent):
    eligible_datatypes = [NumberDataType]

    def prepare(self, datatype, *, context, **kwargs):
        minimum = datatype.definition.get("minimum", None)
        maximum = datatype.definition.get("maximum", None)
        exclusive_minimum = datatype.definition.get("exclusiveMinimum", None)
        exclusive_maximum = datatype.definition.get("exclusiveMaximum", None)

        if (
            minimum is None
            and maximum is None
            and exclusive_minimum is None
            and exclusive_maximum is None
        ):
            return
        m = datatype.definition.setdefault("marshmallow", {})
        validators = m.setdefault("validators", [])
        args = []

        if minimum is not None:
            args.append("min=%s" % minimum)
        if maximum is not None:
            args.append("max=%s" % maximum)
        if exclusive_minimum is not None:
            args.append("min=%s" % exclusive_minimum)
            args.append("min_inclusive=False")
        if exclusive_maximum is not None:
            args.append("max=%s" % exclusive_maximum)
            args.append("max_inclusive=False")
        validators.append("ma.validate.Range(%s)" % ", ".join(args))



from typing import List


def dict_get(data, path: List[str], default=None):
    for p in path[:-1]:
        if not data:
            data = {}
        data = data.get(p)
    return (data or {}).get(path[-1], default)


def dict_setdefault(data, path: List[str], default=None):
    for p in path[:-1]:
        r = data.get(p)
        if not data:
            data.setdefault(p, {})
        data = r
    if path[-1] not in data:
        data[path[-1]] = default
    return data[path[-1]]



import os

import marshmallow as ma

from oarepo_model_builder.datatypes import DataTypeComponent, ModelDataType
from oarepo_model_builder.utils.python_name import module_to_path, parent_module
from oarepo_model_builder.validation.utils import PermissiveSchema

from .defaults import DefaultsModelComponent
from .jsonschema import JSONSchemaModelComponent
from .record import RecordModelComponent
from .utils import set_default


class ModelMappingSchema(ma.Schema):
    generate = ma.fields.Bool(metadata={"doc": "Generate mapping (default is true)"})
    alias = ma.fields.Str(
        metadata={
            "doc": "Index alias, under which the mapping is registered in setup.cfg"
        }
    )
    module = ma.fields.Str(metadata={"doc": "Module with mapping definition"})
    index = ma.fields.Str(metadata={"doc": "Index name"})
    file_ = ma.fields.Str(
        data_key="file", attribute="file", metadata={"doc": "Path to index file"}
    )

    # mapping
    template = ma.fields.Nested(
        PermissiveSchema,
        metadata={"doc": "Mapping template, merged with generated mapping"},
    )
    skip = ma.fields.Boolean()
    index_field_arguments = ma.fields.List(
        ma.fields.String(),
        data_key="index-field-arguments",
        attribute="index-field-arguments",
        metadata={"doc": "Other arguments for index field"},
    )

    class Meta:
        unknown = ma.RAISE


class MappingModelComponent(DataTypeComponent):
    eligible_datatypes = [ModelDataType]
    depends_on = [
        DefaultsModelComponent,
        RecordModelComponent,
        JSONSchemaModelComponent,
    ]

    class ModelSchema(ma.Schema):
        mapping = ma.fields.Nested(
            ModelMappingSchema,
            attribute="mapping",
            data_key="mapping",
            metadata={"doc": "Mapping definition"},
        )
        searchable = ma.fields.Bool(
            load_default=True,
            metadata={
                "doc": "Will the mapping/indexing be generated on model? (can be overriden on individual properties)"
            },
        )

    def before_model_prepare(self, datatype, *, context, **kwargs):
        prefix_snake = datatype.definition["module"]["prefix-snake"]
        alias = datatype.definition["module"]["alias"]
        records_path = module_to_path(
            parent_module(datatype.definition["record"]["module"])
        )

        mapping = set_default(datatype, "mapping", {})
        mapping.setdefault("generate", True)
        alias = mapping.setdefault("alias", alias)
        mapping.setdefault(
            "module",
            f'{parent_module(datatype.definition["record"]["module"])}.mappings',
        )
        short_index_name = (
            f"{prefix_snake}-{datatype.definition['json-schema-settings']['version']}"
        )
        mapping.setdefault(
            "index",
            f"{alias}-{short_index_name}",
        )
        mapping.setdefault(
            "file",
            os.path.join(
                records_path,
                "mappings",
                "os-v2",
                alias,
                f"{short_index_name}.json",
            ),
        )
        mapping.setdefault("index-field-args", [])



import marshmallow as ma

from oarepo_model_builder.datatypes import ArrayDataType, DataTypeComponent
from oarepo_model_builder.validation.utils import StrictSchema


class SampleSchema(StrictSchema):
    skip = ma.fields.Boolean(
        metadata={"doc": "Set true to skip generating sample for the field"}
    )
    faker = ma.fields.String(
        metadata={"doc": "The faker to use for generating the sample"}
    )
    params = ma.fields.Raw(metadata={"doc": "Params for the faker"})

    def load(
        self,
        data,
        *,
        many=None,
        partial=None,
        unknown=None,
    ):
        if isinstance(data, (list, tuple)):
            return data
        return super().load(data, many=many, partial=partial, unknown=unknown)

    def dump(self, obj, *, many=None):
        if many:
            return [self.dump(x, many=False) for x in obj]
        if isinstance(obj, (list, tuple)):
            return obj
        return super().dump(obj, many=False)


class RegularSampleComponent(DataTypeComponent):
    class ModelSchema(ma.Schema):
        sample = ma.fields.Nested(
            SampleSchema,
            required=False,
        )


class ArraySampleSchema(StrictSchema):
    count = ma.fields.Int()


class ArraySampleComponent(RegularSampleComponent):
    eligible_datatypes = [ArrayDataType]

    class ModelSchema(StrictSchema):
        sample = ma.fields.Nested(
            ArraySampleSchema,
            required=False,
        )



import sys
from typing import Mapping

import libcst as cst
from jinja2 import Environment, FunctionLoader, Undefined, pass_context

from oarepo_model_builder.outputs import OutputBase
from oarepo_model_builder.templates import templates
from oarepo_model_builder.utils.cst import PythonContext, merge
from oarepo_model_builder.utils.jinja import (
    base_name,
    class_header,
    generate_dict,
    generate_extra_code,
    generate_extra_code_imports,
    generate_import,
    generate_list,
    in_different_package,
    package_name,
    repr_filter,
)
from oarepo_model_builder.utils.verbose import log


class PythonOutput(OutputBase):
    TYPE = "python"
    cst = None
    original_data = None

    def begin(self):
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.original_data = f.read()

                self.cst = cst.parse_module(self.original_data)
        except FileNotFoundError:
            self.original_data = None
            self.cst = cst.parse_module("")

    @property
    def created(self):
        return self.original_data is None

    def finish(self):
        code = self.cst.code
        if code != self.original_data:
            self.builder.filesystem.mkdir(self.path.parent)
            log(2, "Saving %s", self.path)
            with self.builder.filesystem.open(self.path, mode="w") as f:
                f.write(code)
            if self.builder.schema.settings["python"]["use-isort"]:
                import isort

                config = isort.settings.Config(verbose=False, quiet=True)
                isort.file(self.path, config=config)
            if self.builder.schema.settings["python"]["use-black"]:
                import subprocess

                subprocess.call(["black", "-q", "--preview", str(self.path)])
            if self.builder.schema.settings["python"]["use-autoflake"]:
                import subprocess

                subprocess.call(
                    [
                        "autoflake",
                        "--in-place",
                        "--remove-all-unused-imports",
                        str(self.path),
                    ]
                )

    def merge(self, template_name, context, filters=None):
        # template is a loadable resource
        env = Environment(
            loader=FunctionLoader(
                lambda tn: templates.get_template(tn, context["settings"])
            ),
            autoescape=False,
            undefined=StrictUndefined,
        )
        self.register_default_filters(env)
        for filter_name, filter_func in (filters or {}).items():
            env.filters[filter_name] = filter_func

        try:
            context = {
                **context,
                **{k.replace("-", "_"): v for k, v in list(context.items())},
            }
            rendered = env.get_template(template_name).render(make_attrdict(context))
        except Exception as exc:
            raise RuntimeError(
                f"Error rendering template {template_name}: {str(exc)}"
            ) from exc
        try:
            rendered_cst = cst.parse_module(
                rendered, config=self.cst.config_for_parsing
            )
        except:
            print(rendered, file=sys.stderr)
            raise

        try:
            self.cst = merge(
                PythonContext(rendered_cst),
                self.cst,
                rendered_cst,
            )
        except:
            print(rendered)
            print("====")
            print(self.original_data)
            raise

    @staticmethod
    def register_default_filters(env):
        env.filters["generate_import"] = generate_import
        env.filters["imports"] = generate_import
        env.filters["extra_code"] = generate_extra_code
        env.filters["code_imports"] = generate_extra_code_imports
        env.filters["generate_list"] = generate_list
        env.filters["generate_dict"] = generate_dict
        env.filters["class_header"] = class_header
        env.filters["package_name"] = package_name
        env.filters["base_name"] = pass_context(lambda context, value: base_name(value))
        env.tests["in_different_package"] = pass_context(
            lambda context, value: in_different_package(
                context["current_module"], value
            )
        )
        env.filters["repr"] = repr_filter


class AttrDict(dict):
    def __getattr__(self, item):
        try:
            return self[item]
        except KeyError as e:
            raise AttributeError(f"No key {item}") from e

    def __getitem__(self, item):
        if super().__contains__(item):
            ret = super().__getitem__(item)
        else:
            ret = super().__getitem__(item.replace("_", "-"))

        if isinstance(ret, Mapping) and not isinstance(ret, AttrDict):
            return AttrDict(ret)
        return ret

    def items(self):
        return [(k, self[k]) for k in self]

    def values(self):
        return [self[k] for k in self]

    def __contains__(self, item):
        return super().__contains__(item) or super().__contains__(
            item.replace("_", "-")
        )


class StrictUndefined(Undefined):
    def __str__(self):
        self._fail_with_undefined_error()


def make_attrdict(d):
    d = AttrDict(d)
    return {k: d[k] for k in d}



import marshmallow as ma
from marshmallow import fields

from oarepo_model_builder.datatypes import DataTypeComponent, datatypes
from oarepo_model_builder.utils.facet_helpers import facet_name
from oarepo_model_builder.validation.utils import StrictSchema


class PropertySortSchema(StrictSchema):
    key = fields.String(required=False)
    order = fields.String(required=False)
    arguments = fields.List(fields.String(), required=False)
    validators = fields.List(fields.String(), required=False)


class FieldSortComponent(DataTypeComponent):
    class ModelSchema(ma.Schema):
        sortable = ma.fields.Nested(
            PropertySortSchema,
            required=False,
        )

    def create_sort_options(self, datatype, *, context, **kwargs):
        definition = datatype.section_sortable.config
        if definition:
            key = definition.get("key", facet_name(datatype.path))
            path = datatype.path
            order = definition.get("order", "asc")
            if order == "desc":
                path = "-" + path

            return {key: {"fields": [path]}}

    def after_model_prepare(self, datatype, *, context, **kwargs):
        for node in datatype.deep_iter():
            datatypes.call_components(node, "create_sort_options", context=context)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordMetadataBuilder(InvenioBaseClassPythonBuilder):
    TYPE = "invenio_record_metadata"
    section = "record-metadata"
    template = "record-metadata"



from oarepo_model_builder.datatypes import NestedDataType

from . import FacetDefinition
from .object import ObjectFacetsComponent


class NestedFacetsComponent(ObjectFacetsComponent):
    eligible_datatypes = [NestedDataType]

    def build_facet_definition(
        self,
        datatype,
        facet_definition: FacetDefinition,
    ):
        facet_section = datatype.section_facets.config

        facet_definition.update(facet_section)
        facet_definition.set_field(
            facet_section,
            arguments=[
                f"path = {repr(datatype.path)}",
                f"nested_facet = {facet_definition.field}",
                *facet_section.get("args", []),
            ],
        )
        return super().build_facet_definition(datatype, facet_definition)

    def process_facets(self, datatype, section, **kwargs):
        # container itself does not generate any facets. Need to have it here because facet-class is set
        # and calling field implementation would generate non meaningful facet record here.
        section.config["facets"] = []
